package PSOut;

# (c) 2003-2004
# written by John McPherson to get postscript output for springgraph.pl
# released 28 Oct 2003 under the terms of the GNU Public Licence.
# (yes, I know Americans spell it with an "s")

use Exporter;
@ISA = qw(Exporter);

# default symbols to export
# needed if GD.pm isn't loaded
BEGIN {
    if (!exists(&gdLargeFont)) {
	@EXPORT = qw (&gdLargeFont);
    }
}
use strict;

sub new {
    shift; # class name
    my %obj;
    $obj{'width'}=shift;
    $obj{'height'}=shift;
    $obj{'colour_list'}=[];
    $obj{'line_list'}=[];
    $obj{'string_list'}=[];
    $obj{'arc_list'}=[];
    bless \%obj;
}


# create and return a new colour, given rgb values

sub colorAllocate {
    my $self=shift;

    my @new_colour=(shift, shift, shift); # r,g,b

    # translate from decimal 0 .. 255 into postscript 0.0 .. 1.0

# do this at output time, not now
    #foreach my $colour (@new_colour) {
    #  $colour /= 255;
    #}

    my $colour_listref=$self->{'colour_list'};
    push @$colour_listref, \@new_colour;
    return @$colour_listref -1; # number of this colour
}

sub colorResolve {
    my $self=shift;
    my @colour=(shift, shift, shift); # r,g,b

    # see if this colour is already in our database
    my $colour_listref=$self->{'colour_list'};
    my $counter=@$colour_listref -1;
    while ($counter >= 0) {
	my @test_colour=@{ @$colour_listref[$counter] };

	if ($test_colour[0] == $colour[0]
	    && $test_colour[1] ==  $colour[1]
	    && $test_colour[2] ==  $colour[2]) {
	    return $counter;
	}
	$counter--;
    }
    # shouldn't get here?
    # create new colour
    return &colorAllocate($self,@colour); # is this correct behaviour?
}

sub line {
    my $self=shift;
    my ($fromx, $fromy, $tox, $toy)=(shift,shift,shift,shift);
    push @{$self->{'line_list'}}, [$fromx, $fromy, $tox, $toy];
    
}



# colour is always set to black? (assume this is for outline colour)
sub arc {
    my $self=shift;

    #$im->arc($point[3]{x},$point[3]{y},10,10,0,360,$black);
    my ($x,$y, $u1, $u2, $u3, $angle, $colour_num) = @_;
    # $u1 is width, $u2 is height (?). Height is always 40?
    push (@{$self->{'arc_list'}},
	  [$x, $y, $u1, $u2, $u3, $angle, $colour_num]);

}



sub string {
    my $self=shift;
    my ($x, $y, $label, $colour_num)=(shift, shift, shift, shift);

    $label =~ s@\(@\\\(@g;
    $label =~ s@\)@\\\)@g;


    # assume text is already in latin1
    #while ($label =~ s@([\x80-\xff])@????($1)@eg) {;}
	
    push @{$self->{'string_list'}}, [$x, $y, $label, $colour_num];
}



sub transparent {

}


# postscript can't easily just take a point and fill.
# but here at least, we are always given the same point as the centre of a
# circle, so we'll just find the corresponding circle and set its colour
sub fillToBorder {
    my $self=shift;
    my ($fillx, $filly, $bordercol, $fillcol)=@_;
    my @arc_list=@{$self->{'arc_list'}};
    foreach my $arcdata (@arc_list) {
	my ($x, $y)=@$arcdata; 
	if ($fillx == $x && $filly == $y) {
	    @$arcdata[6]=$fillcol;
	    return;
	}
    }
    # shouldn't get here
    print STDERR "warning: given fill but can't find corresponding circle!\n";
}

sub gdLargeFont {}

sub print_ps {
    my $self=shift;

    my @colour_list=@{$self->{'colour_list'}};
    my @string_list=@{$self->{'string_list'}};
    my @arc_list=@{$self->{'arc_list'}};
    my @line_list=@{$self->{'line_list'}};

    # print postscript header
    print "%!PS-Adobe-2.0\n%%Version: 1 2\n%%LanguageLevel: 2\n";
    # do we want to allow the title to be customisable?
    print "%%Pages: 1\n%%Orientation: Portrait\n";
    print "%%Title: GPG Key signature graph\n";
    print "%%BoundingBox: 0 0 ". $self->{'width'}." ". $self->{'height'} ."\n";
    print "%%EndComments\n";

    # font encoding issues. use iso-8859-1
    # print "/Times-Roman findfont 15 scalefont setfont\n"; # doesn't work

    # example given in postscript documentation from adobe
    print "/Times-Roman findfont dup length dict begin {1 index /FID ne\n";
    print "{def} {pop pop} ifelse} forall /Encoding ISOLatin1Encoding def\n";
    print "currentdict end dup /FontName get 80 string cvs (-ISOLatin1)\n";
    print "concatstrings cvn exch definefont\n\n";
    print "15 scalefont setfont\n";


    # create a function for drawing circle and text
    my $blackcolnum=colorResolve($self, 0,0,0);
    # exch swaps the top 2 items on the stack
    print "/DrawOval {\n% parameters\n /xscale exch def\n /y exch def\n";
    print " /x exch def\n";
    print " save % save current state\n 0.5 setlinewidth\n";
    print " x y translate % move by this offset\n";
    print " xscale 1 scale % changes ALL measurements\n";
    print " 0 0 20  0 360 arc fill\n"; # 20 is the original radius
    print " col$blackcolnum 0 0 20  0 360 arc stroke\n";
    print " restore % restore state\n} bind def\n\n";

    # assuming that the string is on the stack, draw a white box at
    # the given x,y position
    my $whitecolnum=colorResolve($self, 255,255,255);

    print "/DrawTextBG {\n/y exch def /x exch def % params\n";
    print " x y moveto\n .4 .4 .75 setrgbcolor\n";
    # use 0.5 setgray instead of colour?

    # define to 0 to use a shadow, define to 1 for a big square white box
    my $use_box_background=0;

    # give the text a white background, for readability
    if ($use_box_background) {
	print " dup false charpath pathbbox % gives corner coordinates\n";
	print " /top exch def /right exch def /bottom exch def /left exch def";
	print "\n newpath left bottom moveto left top lineto";
	print "\n right top lineto right bottom lineto closepath fill\n";
    } else {
	# shadow
	print " x .8 add y .8 sub moveto\n";
	print " dup show\n"; # duplicate the string
#	print " dup x .5 sub y .5 add moveto show\n";
    }
    print "} bind def \n\n";
    # set up our colour definitions
    my $counter=$#colour_list;
    while ($counter >= 0) {
	my $colour_ref=$colour_list[$counter];
	my ($r, $g, $b)=(@$colour_ref);
	foreach my $col ($r,$g,$b) {
	    $col = sprintf("%0.6f", $col / 255);
	}
	print "/col$counter {$r $g $b setrgbcolor} bind def\n";
	$counter--;
    }

    # page data
    print "%%Page: 1 1\n0 0 0 setrgbcolor\n1.2 setlinewidth\n"; # set the pen to black


    # print out the lines
    print "\n% line data\n";
    foreach my $linedata (@line_list) {
	my ($fromx, $fromy, $tox, $toy)=@$linedata; 
	print "$fromx $fromy moveto $tox $toy lineto stroke\n";
    }


    # print out the circles/arcs
    print "\n% oval data\n";
    foreach my $arcdata (@arc_list) {
	my ($x, $y, $width, $height, $u3, $angle, $colour_num)=@$arcdata; 
	# u1 and u2 are width and height?
	# radius is $height/2 - assume(!) height is always 40
	my $xscale=$width/40;
	print " col$colour_num $x $y $xscale DrawOval\n";

#
# $x $y translate $xstretch $ystretch scale
#  xoffset? yoffset? size startangle endangle arc
# 4 1 scale % stretch 4x wide
# 100 400 translate % hrm.... in what units?
# 200 200 400 0 360 arc % units?
    }


    # print out the labels
    print "\n% label/strings data\n";

    foreach my $stringdata (@string_list) {
	my ($fromx, $fromy, $string, $colnum)=@$stringdata; 

	print "($string) $fromx $fromy DrawTextBG\n";
	print "col$colnum $fromx $fromy moveto show\n"; # string still on stack
    }

    # footer

    print "%%EndPage\nshowpage\n%%Trailer\n%%EOF\n";
}

1;
